/*
** pork.c - main function.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <pork.h>

struct screen screen;

/*
** The fallback for when no binding for a key exists:
**
** Write the key to the input buffer.  If the key is a control
** character, output ^<KEY> to avoid messing up the display.
*/

static void binding_insert(int key) {
	struct imwindow *imwindow = screen.cur_window->data;

	if (key <= 0xff) {
		if (iscntrl(key)) {
			static char buf[3] = { '^', '\0', '\0' };

			buf[1] = toupper(key) + 'A' - 1;
			input_insert_str(imwindow->input, buf);
		} else
			input_insert(imwindow->input, key);
	}
}

static inline void binding_run(struct binding *binding) {
	run_command(binding->binding, CMDSET_MAIN);
}

static void sigwinch_handler(int sig __notused) {
	struct winsize size;
	struct imwindow *imwindow = screen.cur_window->data;

	if (ioctl(1, TIOCGWINSZ, &size) != 0)
		pork_exit(-1, "Fatal error getting screen size\n");

	if (size.ws_row < 3 || size.ws_col < 2)
		pork_exit(-1, "You made the screen too small.  Don't do that\n");

	resizeterm(size.ws_row, size.ws_col);
	screen_resize(&screen, size.ws_row, size.ws_col);
	status_draw(&screen, imwindow->priv);
	screen_refresh(&screen);
}

static void sigsegv_handler(int sig __notused) {
	endwin();
	fprintf(stderr, "You found a bug.\n");
	exit(-1);
}

static void generic_signal_handler(int sig) {
	pork_exit(sig, "Caught signal %d.  Exiting\n", sig);
}

static void keyboard_input(void) {
	struct imwindow *imwindow = screen.cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	int key;

	key = wgetinput(screen.status_bar);
	if (key == -1)
		return;

	time(&priv->last_input);

	if (priv->connected && priv->marked_idle) {
		screen_status_msg("You're no longer marked as idle");
		aim_set_idle(priv, 0);
		priv->marked_idle = 0;
	}

	bind_exec(imwindow->active_binds, key);
	imwindow = screen.cur_window->data;

	if (imwindow->blist_visible)
		blist_refresh(imwindow->priv->blist, imwindow);

	status_draw(&screen, imwindow->priv);
	screen_redraw_input(&screen);

	screen_doupdate();
}

static void pork_update_acct(void *data, void *nothing __notused) {
	struct aim_acct *acct = data;
	struct aim_priv *priv = acct->priv;

	if (priv->connected) {
		time_t time_now = time(NULL);

		/*
		** We've got to keep track of buddies' idle time
		** ourselves; the aim server reports an idle time
		** for them only when there's a change (i.e. they
		** go idle, come back from idle or their client reports
		** a new idle time).  This routine will run once a minute
		** and increment the idle time of buddies where appropriate.
		*/

		if (time_now >= priv->last_update + 60) {
			struct imwindow *imwindow = screen.cur_window->data;

			aim_keepalive(priv);
			buddy_update_idle(priv);
			priv->last_update = time_now;

			/*
			** Only refresh the buddy list if it's displayed
			** in the currently visible window (it could be
			** the buddy list of an account that's not bound
			** to the current window.
			*/

			if (imwindow->priv == priv) {
				if (!imwindow_refresh_blist(screen.cur_window->data))
					screen_redraw_input(&screen);
			}
		}

		if (priv->buddy_pref != NULL && priv->buddy_pref->report_idle &&
			!priv->marked_idle)
		{
			time_t time_diff = time_now - priv->last_input;
			int idle_after = opt_get_int(OPT_IDLE_AFTER);

			if (idle_after > 0 && time_diff >= 60 * idle_after) {
				screen_status_msg("Marking you idle after %d minutes of inactivity",
					time_diff / 60);
				aim_set_idle(priv, time_diff);
			}
		}
	}
}

int main(int argc, char **argv) {
	struct imwindow *imwindow;
	int ret;

	if (get_options(argc, argv) != 0) {
		fprintf(stderr, "Fatal: Error getting options.\n");
		exit(-1);
	}

	if (initialize_environment() != 0) {
		fprintf(stderr, "Fatal: Error initializing the terminal.\n");
		exit(-1);
	}

	init_colors();
	if (screen_init(&screen, LINES, COLS) == -1)
		pork_exit(0, "Fatal: Error initializing the terminal.\n");

	signal(SIGWINCH, sigwinch_handler);
	signal(SIGTERM, generic_signal_handler);
	signal(SIGSEGV, sigsegv_handler);

	wmove(screen.status_bar, STATUS_ROWS - 1, 0);
	imwindow = screen.cur_window->data;

	bind_init(&screen.binds);
	bind_set_handlers(&screen.binds.main, binding_run, binding_insert);
	bind_set_handlers(&screen.binds.blist, binding_run, NULL);
	alias_init(&screen.alias_hash);

	screen_set_quiet(&screen, 1);
	ret = read_global_config();
	screen_set_quiet(&screen, 0);

	if (ret != 0)
		screen_status_msg("Error reading the global configuration.");

	status_draw(&screen, &screen.null_priv);
	screen_redraw_input(&screen);
	screen_doupdate();

	while (1) {
		struct timeval tv = { 0, 500000 };
		int ret;
		aim_session_t *session;
		aim_conn_t *cur_conn = NULL;

		session = aim_select_all(screen.acct_list, &tv, &ret, &cur_conn);

		if (ret == -1) {
			if (session != NULL)
				aim_disconnected(session->aux_data);
		} else if (ret == 1) {
			/* Outgoing */
			if (session != NULL) {
				aim_tx_flushqueue(session);
				screen_redraw_input(&screen);
				screen_doupdate();
			}
		} else if (ret == 2) {
			/* Incoming */
			if (cur_conn->fd == STDIN_FILENO) {
				keyboard_input();
			} else {
				if (session != NULL && cur_conn != NULL &&
					aim_get_command(session, cur_conn) >= 0)
				{
					aim_rxdispatch(session);
					imwindow = screen.cur_window->data;

					imwindow_refresh_blist(imwindow);
					status_draw(&screen, imwindow->priv);
					screen_redraw_input(&screen);
					screen_doupdate();
				}
			}
		}

		dlist_iterate(screen.acct_list, pork_update_acct, NULL);
	}

	pork_exit(0, NULL);
}

/*
** Sign off all acounts and exit.  If a message
** is given, print it to the screen.
*/

void pork_exit(int status, char *fmt, ...) {
	dlist_t *cur = screen.acct_list;

	while (cur != NULL) {
		struct aim_acct *acct = cur->data;
		dlist_t *next = cur->next;

		screen_del_acct(&screen, acct->username);
		cur = next;
	}

	endwin();

	if (fmt != NULL) {
		va_list va;

		va_start(va, fmt);
		vfprintf(stderr, fmt, va);
		va_end(va);
	}

	exit(status);
}

/*
** Initialize an "aim_priv" structure for the
** account owned by the specified user.
** Each user account gets its own priv structure,
** which contains account specific data.
*/

struct aim_priv *pork_priv_init(const char *user) {
	struct aim_priv *priv;

	priv = xcalloc(1, sizeof(*priv));
	priv->username = xstrdup(user);

	if (buddy_init(priv) != 0) {
		free(priv->username);
		free(priv);
		return (NULL);
	}

	read_user_config(priv, priv->username);
	blist_init(priv);

	aim_setup(priv);
	time(&priv->last_input);
	return (priv);
}

/*
** Free everything that needs freeing in the
** given "aim_priv" structure.
*/

void pork_priv_free(struct aim_priv *priv) {
	aim_kill_all_conn(priv);

	save_user_config(priv, priv->username);

	blist_destroy(priv->blist);
	buddy_destroy(priv);

	if (priv->passwd != NULL) {
		memset(priv->passwd, 0, strlen(priv->passwd));
		free(priv->passwd);
	}

	free(priv->username);
	free(priv->proxy);
	free(priv->proxy_user);
	free(priv->proxy_passwd);
	free(priv->server);
	free(priv->profile);
	free(priv->away_msg);

	free(priv);
}
