/*
** pork_aim.c - Interface to libfaim.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This file is heavily based on src/protocols/oscar/oscar.c from gaim,
** which is more or less to say that chunks of code have been mostly
** copied from there.  Any copyrights there apply here.
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>

#include <pork.h>

static u_int32_t pork_caps = AIM_CAPS_CHAT;

static char *msgerrreason[] = {
	"Invalid error",
	"Invalid SNAC",
	"Rate to host",
	"Rate to client",
	"Not logged in",
	"Service unavailable",
	"Service not defined",
	"Obsolete SNAC",
	"Not supported by host",
	"Not supported by client",
	"Refused by client",
	"Reply too big",
	"Responses lost",
	"Request denied",
	"Busted SNAC payload",
	"Insufficient rights",
	"In local permit/deny",
	"Sender's warning level is too high",
	"Receiver's warning level is too high",
	"User temporarily unavailable",
	"No match",
	"List overflow",
	"Request ambiguous",
	"Queue full",
	"Not while on AOL"
};

void aim_debug(	aim_session_t *session __notused,
				int level __notused,
				const char *format __notused,
				va_list va __notused)
{
#ifdef ENABLE_DEBUGGING
	char buf[8192];

	vsnprintf(buf, sizeof(buf), format, va);
	screen_status_msg("LIBFAIM DEBUG: lev: %d: %s", level, buf);
#endif
}

static FAIM_CB(aim_recv_bos_rights) {
	va_list ap;
	struct aim_priv *priv = session->aux_data;

	priv->bos_conn = fr->conn;

	va_start(ap, fr);
	priv->rights.max_permit = va_arg(ap, unsigned int);
	priv->rights.max_deny = va_arg(ap, unsigned int);
	va_end(ap);

#if 0
	/* not yet */
	aim_reqservice(session, fr->conn, AIM_CONN_TYPE_CHATNAV);
	aim_reqservice(session, fr->conn, AIM_CONN_TYPE_EMAIL);
#endif

	aim_ssi_reqrights(session, fr->conn);
	aim_ssi_reqdata(session, fr->conn, session->ssi.timestamp,
		session->ssi.revision);

	return (1);
}

static FAIM_CB(aim_recv_buddy_rights) {
	va_list ap;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	priv->rights.max_buddies = va_arg(ap, unsigned int);
	priv->rights.max_watchers = va_arg(ap, unsigned int);
	va_end(ap);

	return (1);
}

static FAIM_CB(aim_recv_conn_complete) {
	aim_reqpersonalinfo(session, fr->conn);
	aim_bos_reqlocaterights(session, fr->conn);
	aim_bos_reqbuddyrights(session, fr->conn);
	aim_reqicbmparams(session);
	aim_bos_reqrights(session, fr->conn);

	return (1);
}

static FAIM_CB(aim_recv_typing) {
	va_list ap;
	u_int16_t type1;
	u_int16_t type2;
	char *source;
	struct imwindow *conv_window;

	va_start(ap, fr);
	type1 = (u_int16_t) va_arg(ap, unsigned int);
	source = va_arg(ap, char *);
	type2 = (u_int16_t) va_arg(ap, unsigned int);
	va_end(ap);

	conv_window = imwindow_find(source);
	if (conv_window == NULL)
		return (0);

	if (type2 > 2)
		return (0);

	conv_window->typing = type2;
	return (0);
}

static FAIM_CB(aim_recv_err_loc) {
	va_list ap;
	char *dest;
	char *err_str;
	u_int16_t code;

	va_start(ap, fr);
	code = (u_int16_t) va_arg(ap, unsigned int);
	dest = va_arg(ap, char *);
	va_end(ap);

	if (code < array_elem(msgerrreason))
		err_str = msgerrreason[code];
	else
		err_str = "Reason unknown";

	screen_status_msg("User information for is %s unavailable: %s",
		dest, err_str);

	return (1);
}

static FAIM_CB(aim_recv_err_msg) {
	va_list ap;
	char *dest;
	char *err_str;
	u_int16_t code;

	va_start(ap, fr);
	code = (u_int16_t) va_arg(ap, unsigned int);
	dest = va_arg(ap, char *);
	va_end(ap);

	if (code < array_elem(msgerrreason))
		err_str = msgerrreason[code];
	else
		err_str = "Reason unknown";

	screen_buddy_status_msg(dest, 0,
		"Your message to %s was not sent: %s", dest, err_str);
	return (1);
}

static FAIM_CB(aim_recv_err_other) {
	va_list ap;
	u_int16_t code;
	char *err_str;

	va_start(ap, fr);
	code = (u_int16_t) va_arg(ap, unsigned int);
	va_end(ap);

	if (code < array_elem(msgerrreason))
		err_str = msgerrreason[code];
	else
		err_str = "Reason unknown";

	screen_status_msg("AIM (SNAC) Error: %s", err_str);
	return (1);
}

static FAIM_CB(aim_recv_evil) {
	va_list ap;
	u_int16_t warn_level;
	aim_userinfo_t *userinfo;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	warn_level = va_arg(ap, unsigned int);
	userinfo = va_arg(ap, aim_userinfo_t *);
	va_end(ap);

	warn_level /= 10;
	if (warn_level < priv->warn_level) {
		priv->warn_level = warn_level;
		return (1);
	}

	if (blank_str(userinfo->sn)) {
		screen_status_msg(
			"You have been warned anonymously.  Your new warning level is %u%%",
			warn_level);
	} else {
		screen_buddy_status_msg(userinfo->sn, 1,
			"You have been warned by %s.  Your new warning level is %u%%",
			userinfo->sn, warn_level);
	}

	priv->warn_level = warn_level;
	return (1);
}

static FAIM_CB(aim_recv_icbm_param_info) {
	va_list ap;
	struct aim_icbmparameters *params;

	va_start(ap, fr);
	params = va_arg(ap, struct aim_icbmparameters *);
	va_end(ap);

	params->flags = 0x0000000b;
	params->maxmsglen = 8000;
	params->minmsginterval = 0;

	aim_seticbmparam(session, params);
	return (1);
}


static int parse_im_chan1(	aim_session_t *session,
							aim_conn_t *conn,
							aim_userinfo_t *userinfo,
							struct aim_incomingim_ch1_args *args)
{
	u_int32_t flags = 0;
	char *buf;
	char *final_msg;
	struct imwindow *conv_window;
	struct aim_priv *priv = session->aux_data;

	(void) session;
	(void) conn;

	if (args->icbmflags & AIM_IMFLAGS_AWAY)
		flags |= 1;

	if (args->icbmflags & AIM_IMFLAGS_UNICODE) {
		int i;
		char *msg;

		buf = xmalloc(args->msglen + 1);
		msg = buf;

		for (i = 0 ; i < args->msglen ; i += 2) {
			u_int16_t uc_char;

			uc_char = ((args->msg[i] & 0xff) << 8) | (args->msg[i + 1] & 0xff);

			if (uc_char > 255)
				*msg++ = (char) uc_char;
			else
				*msg++ = '?';
		}

		*msg = '\0';
	} else
		buf = xstrdup(args->msg);

	final_msg = strip_html(buf);
	free(buf);

	conv_window = imwindow_find(userinfo->sn);
	if (conv_window == NULL) {
		conv_window = screen_new_window(&screen, priv, userinfo->sn);
		if (conv_window == NULL)
			return (-1);
	}

	conv_window->typing = 0;
	imwindow_recv_im(conv_window, userinfo->sn, final_msg);
	free(final_msg);

	return (1);
}

static FAIM_CB(aim_recv_msg) {
	va_list ap;
	int channel;
	aim_userinfo_t *userinfo;
	int ret = 0;

	va_start(ap, fr);
	channel = va_arg(ap, int);
	userinfo = va_arg(ap, aim_userinfo_t *);

	switch (channel) {
		case 1:
		{
			struct aim_incomingim_ch1_args *args;

			args = va_arg(ap, struct aim_incomingim_ch1_args *);
			ret = parse_im_chan1(session, fr->conn, userinfo, args);
		}
		break;

		/* XXX - fix this for directIM */

	}

	va_end(ap);
	return (ret);
}

static FAIM_CB(aim_recv_locrights) {
	va_list ap;
	u_int16_t max_len;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	max_len = va_arg(ap, int);
	va_end(ap);

	priv->rights.max_profile_len = max_len;
	priv->rights.max_away_len = max_len;

	aim_bos_setprofile(session, fr->conn, priv->profile, NULL, pork_caps);
	return (1);
}

static FAIM_CB(aim_recv_mem_request) {
	screen_status_msg("Event: aim_recv_mem_request");
	return (1);
}

static FAIM_CB(aim_recv_clientauto) {
	/* XXX - ignore this for now */
	return (1);
}

static FAIM_CB(aim_recv_missed) {
	va_list ap;
	u_int16_t chan;
	u_int16_t num_missed;
	u_int16_t reason;
	aim_userinfo_t *userinfo;
	char buf[1024];

	va_start(ap, fr);
	chan = (u_int16_t) va_arg(ap, unsigned int);
	userinfo = va_arg(ap, aim_userinfo_t *);
	num_missed = (u_int16_t) va_arg(ap, unsigned int);
	reason = (u_int16_t) va_arg(ap, unsigned int);
	va_end(ap);

	switch(reason) {
		case 0:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"You missed %d message from %s because it was invalid" :
					"You missed %d messages from %s because they were invalid",
				num_missed, userinfo->sn);
			break;

		case 1:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"You missed %d message from %s because it was too large" :
					"You missed %d messages from %s because they were too large",
				num_missed,
				userinfo->sn);
			break;

		case 2:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"You missed %d message from %s because the rate limit has been exceeded" :
					"You missed %d messages from %s because the rate limit has been exceeded",
				num_missed, userinfo->sn);
			break;

		case 3:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"You missed %d message from %s because the sender's warning level is too high" :
					"You missed %d messages from %s because the sender's warning level is too high",
				num_missed, userinfo->sn);
			break;

		case 4:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"You missed %d message from %s because your warning level is too high" :
					"You missed %d messages from %s because your warning level is too high",
				num_missed, userinfo->sn);
			break;

		default:
			snprintf(buf, sizeof(buf),
				num_missed == 1 ?
					"You missed %d message from %s for unknown reasons" :
					"You missed %d messages from %s for unknown reasons",
				num_missed, userinfo->sn);
			break;
	}

	return (1);
}

static FAIM_CB(aim_recv_selfinfo) {
	va_list ap;
	aim_userinfo_t *info;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	info = va_arg(ap, aim_userinfo_t *);
	va_end(ap);

	priv->warn_level = (float) info->warnlevel / 10;
	priv->idle_time = info->idletime;

	return (1);
}

static FAIM_CB(aim_recv_offgoing) {
	va_list ap;
	aim_userinfo_t *userinfo;
	struct buddy *buddy;
	struct aim_priv *priv = session->aux_data;
	struct imwindow *conv_window;

	va_start(ap, fr);
	userinfo = va_arg(ap, aim_userinfo_t *);
	va_end(ap);

	if (userinfo == NULL)
		return (-1);

	buddy = buddy_find(priv, userinfo->sn);
	if (buddy == NULL) {
		/* XXX - a buddy who's not on our list? */
		return (1);
	}

	/*
	** If there's a conversation window with this user,
	** print the signoff message inside that window.
	**
	** If there's no conversation window, but notify
	** is set, print the signoff message in the current window.
	*/

	conv_window = imwindow_find(userinfo->sn);
	if (conv_window != NULL) {
		screen_buddy_status_msg(userinfo->sn, 1, "%s has signed off",
			userinfo->sn);
	} else {
		struct buddy *buddy = buddy_find(priv, userinfo->sn);

		if (buddy != NULL && buddy->notify) {
			screen_buddy_status_msg(userinfo->sn, 1, "%s has signed off",
				userinfo->sn);
		}
	}

	buddy_offline(priv, buddy);
	return (1);
}

static FAIM_CB(aim_recv_oncoming) {
	va_list ap;
	aim_userinfo_t *userinfo;
	struct buddy *buddy;
	struct aim_priv *priv = session->aux_data;
	char *status_msg;

	va_start(ap, fr);
	userinfo = va_arg(ap, aim_userinfo_t *);
	va_end(ap);

	if (userinfo == NULL)
		return (-1);

	buddy = buddy_find(priv, userinfo->sn);
	if (buddy == NULL) {
		/* XXX - a buddy who's not on our list? */
		return (1);
	}

	if (buddy->status == STATUS_OFFLINE) {
		status_msg = "has signed on";
		buddy_online(priv, buddy, userinfo);
	} else {
		int was_away = buddy->status == STATUS_AWAY;
		int is_away = (userinfo->flags & AIM_FLAG_AWAY) != 0;

		if (was_away != is_away) {
			if (was_away)
				status_msg = "is no longer away";
			else
				status_msg = "is now away";
		} else if (buddy->idle_time && !userinfo->idletime)
			status_msg = "is no longer idle";
		else if (!buddy->idle_time && userinfo->idletime)
			status_msg = "is now idle";
		else
			status_msg = NULL;

		buddy_update(priv, buddy, userinfo);
	}

	/*
	** If there's a conversation window with this user,
	** print the signoff message inside that window.
	**
	** If there's no conversation window, but notify
	** is set, print the signoff message in the current window.
	*/

	if (status_msg != NULL) {
		struct imwindow *conv_window = imwindow_find(userinfo->sn);
		if (conv_window != NULL) {
			screen_buddy_status_msg(userinfo->sn, 1, "%s %s",
				userinfo->sn, status_msg);
		} else {
			struct buddy *buddy = buddy_find(priv, userinfo->sn);

			if (buddy != NULL && buddy->notify) {
				screen_buddy_status_msg(userinfo->sn, 1, "%s %s",
					userinfo->sn, status_msg);
			}
		}
	}

	return (1);
}

static FAIM_CB(aim_recv_rate_change) {
	va_list ap;
	u_int16_t rate_code;
	u_int16_t rate_class;
	u_int32_t window_size;
	u_int32_t rate_clear;
	u_int32_t alert;
	u_int32_t limit;
	u_int32_t disconnect;
	u_int32_t current_avg;
	u_int32_t max_avg;

	va_start(ap, fr);

	rate_code = (u_int16_t) va_arg(ap, unsigned int);
	rate_class = (u_int16_t) va_arg(ap, unsigned int);
	window_size = (u_int32_t) va_arg(ap, unsigned long);
	rate_clear = (u_int32_t) va_arg(ap, unsigned long);
	alert = (u_int32_t) va_arg(ap, unsigned long);
	limit = (u_int32_t) va_arg(ap, unsigned long);
	disconnect = (u_int32_t) va_arg(ap, unsigned long);
	current_avg = (u_int32_t) va_arg(ap, unsigned long);
	max_avg = (u_int32_t) va_arg(ap, unsigned long);

	va_end(ap);

	switch (rate_code) {
		case AIM_RATE_CODE_CHANGE:
			if (current_avg >= rate_clear)
				aim_conn_setlatency(fr->conn, 0);
			break;

		case AIM_RATE_CODE_WARNING:
			aim_conn_setlatency(fr->conn, window_size / 4);
			break;

		case AIM_RATE_CODE_LIMIT:
			aim_conn_setlatency(fr->conn, window_size / 2);
			screen_status_msg("The last message was not sent because you are over the rate limit.  Please wait 10 seconds, and then try again");
			break;

		case AIM_RATE_CODE_CLEARLIMIT:
			aim_conn_setlatency(fr->conn, 0);
			break;
	}

	return (1);
}

static FAIM_CB(aim_recv_redirect) {
	screen_status_msg("Event: aim_recv_redirect");
	return (1);
}

static FAIM_CB(aim_recv_search_error) {
	va_list ap;
	char *address;

	va_start(ap, fr);
	address = va_arg(ap, char *);
	va_end(ap);

	screen_status_msg("No results were found for %s", address);
	return (1);
}

static FAIM_CB(aim_recv_search_reply) {
	va_list ap;
	char *address;
	char *usernames;
	char buf[2048];
	int i = 0;
	size_t len = sizeof(buf);
	int j;
	int num;
	int ret;

	va_start(ap, fr);
	address = va_arg(ap, char *);
	num = va_arg(ap, int);
	usernames = va_arg(ap, char *);
	va_end(ap);

	if (num < 1) {
		screen_status_msg("No results were found for %s", address);
		return (1);
	}

	ret = snprintf(buf, len, "%s has registered the following screen names: %s",
			address, &usernames[0]);
	len -= ret;
	i += ret;

	for (j = 1 ; j < num ; j++) {
		ret = snprintf(&buf[i], len, ", %s", &usernames[j * (MAXSNLEN + 1)]);
		len -= ret;
		i += ret;
	}

	screen_status_msg("%s", buf);
	return (1);
}

/*
** Nothing to do here for now.
*/

static FAIM_CB(aim_ssi_recv_rights) {
	(void) session;
	(void) fr;

	return (1);
}

static FAIM_CB(aim_ssi_recv_list) {
	struct aim_ssi_item *cur;
	struct aim_priv *priv = session->aux_data;

	/*
	** I'm not sure why this would happen more than once.
	** But it does.
	*/
	if (priv->connected)
		return (1);

	aim_ssi_enable(session, fr->conn);
	aim_ssi_cleanlist(session, fr->conn);

	/*
	** Merge the buddy list on the server with the locally
	** stored buddy list.  By this point, the local
	** buddy list will have already been processed.
	*/

	for (cur = session->ssi.items ; cur != NULL ; cur = cur->next) {
		if (cur->type == 0) {
			struct buddy *buddy;
			struct aim_ssi_item *ssi_group;

			if (cur->name == NULL)
				continue;

			buddy = buddy_find(priv, cur->name);
			if (buddy != NULL)
				continue;

			ssi_group = session->ssi.items;
			for (; ssi_group != NULL ; ssi_group = ssi_group->next) {
				struct group *group;

				if (ssi_group->name == NULL)
					continue;

				if (ssi_group->type != 1)
					continue;

				if (ssi_group->gid != cur->gid)
					continue;

				group = group_add(priv, ssi_group->name, 0);
				buddy = buddy_add(priv, cur->name, group, 0);
				buddy_alias(priv, buddy, cur->name);
			}
		} else if (cur->type == 2) {
			if (cur->name == NULL)
				continue;

			buddy_add_permit(priv, cur->name, 0);
		} else if (cur->type == 3) {
			/* Deny */
			if (cur->name == NULL)
				continue;

			buddy_add_block(priv, cur->name, 0);
		} else if (cur->type == 4) {
			u_int32_t permdeny_mode = aim_ssi_getpermdeny(session->ssi.items);

			priv->buddy_pref->permdeny_mode = permdeny_mode;
		} else if (cur->type == 5) {
			u_int32_t report_idle = aim_ssi_getpresence(session->ssi.items);

			if (report_idle & 0x0000400)
				priv->buddy_pref->report_idle = 1;
			else
				priv->buddy_pref->report_idle = 0;
		}
	}

	aim_report_idle(priv, priv->buddy_pref->report_idle);

	aim_ssi_setpermdeny(session, fr->conn,
		priv->buddy_pref->permdeny_mode, 0xffffffff);

	aim_send_buddy_list(session, fr->conn, priv);
	aim_send_permit_list(session, fr->conn, priv);
	aim_send_block_list(session, fr->conn, priv);

	aim_clientready(session, fr->conn);

	priv->connected = 1;
	time(&priv->last_update);
	screen_status_msg("Logged in as %s successfully", priv->username);

	if (priv->profile == NULL)
		aim_set_profile(priv, DEFAULT_AIM_PROFILE);

	return (1);
}

static FAIM_CB(aim_recv_popup) {
	/* Ignore this for now. */
	return (1);
}

static FAIM_CB(aim_recv_userinfo) {
	struct aim_priv *priv = session->aux_data;
	aim_userinfo_t *userinfo;
	u_int16_t info_type;
	char *encoding;
	char *result;
	va_list ap;
	char *profile = NULL;
	char *away_msg = NULL;
	u_int32_t idle_time = 0;
	u_int32_t online_since = 0;
	u_int32_t member_since = 0;
	int print_all = 1;

	va_start(ap, fr);
	userinfo = va_arg(ap, aim_userinfo_t *);
	info_type = (u_int16_t) va_arg(ap, unsigned int);
	encoding = va_arg(ap, char *);
	result = va_arg(ap, char *);
	va_end(ap);

	if (userinfo->present & AIM_USERINFO_PRESENT_ONLINESINCE)
		online_since = userinfo->onlinesince;

	if (userinfo->present & AIM_USERINFO_PRESENT_MEMBERSINCE)
		member_since = userinfo->membersince;

	if (userinfo->present & AIM_USERINFO_PRESENT_IDLE)
		idle_time = userinfo->idletime;

	switch (info_type) {
		case AIM_GETINFO_AWAYMESSAGE:
			if (result != NULL && *result != '\0')
				away_msg = strip_html(result);

			break;

		case AIM_GETINFO_CAPABILITIES:
			break;

		default:
			if (result != NULL && *result != '\0')
				profile = strip_html(result);

			break;
	}

	if (info_type == AIM_GETINFO_AWAYMESSAGE && priv->whois_hack) {
		print_all = 0;
		priv->whois_hack = 0;
	}

	screen_print_info(userinfo->sn, profile, away_msg,
		idle_time, online_since, member_since, print_all);

	if (away_msg != NULL)
		free(away_msg);

	if (profile != NULL)
		free(profile);

	return (1);
}

static FAIM_CB(aim_parse_login) {
	struct aim_priv *priv = (struct aim_priv *) session->aux_data;
	char *key;
	va_list ap;
	static struct client_info_s client_info = CLIENTINFO_AIM_KNOWNGOOD;

	va_start(ap, fr);
	key = va_arg(ap, char *);
	va_end(ap);

	aim_send_login(session, fr->conn,
		priv->username, priv->passwd, &client_info, key);

	return (1);
}

static FAIM_CB(aim_parse_authresp) {
	va_list ap;
	struct aim_authresp_info *authresp;
	aim_conn_t *bos_conn;
	struct aim_priv *priv = session->aux_data;

	va_start(ap, fr);
	authresp = va_arg(ap, struct aim_authresp_info *);
	va_end(ap);

	if (authresp->errorcode || !authresp->bosip || !authresp->cookie) {
		switch (authresp->errorcode) {
			/* Incorrect nick / passwd */
			case 0x05:
				screen_status_msg("Error logging in: Incorrect username or password");
				if (priv->passwd != NULL) {
					memset(priv->passwd, 0, strlen(priv->passwd));
					free(priv->passwd);
					priv->passwd = NULL;
				}
				break;

			/* Suspended */
			case 0x11:
				screen_status_msg("Error logging in: This account has been suspended");
				break;

			/* connecting too frequently */
			case 0x18:
				screen_status_msg("Error logging in: You have been connecting too frequently");
				break;

			/* client too old */
			case 0x1c:
				screen_status_msg("Error logging in: The AIM server thinks this client is too old");
				break;
		}

		aim_conn_kill(session, &fr->conn);
		return (1);
	}

	aim_conn_kill(session, &fr->conn);

	bos_conn = aim_newconn(session, AIM_CONN_TYPE_BOS, authresp->bosip);
	if (bos_conn == NULL)
		return (0);

	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_CONNINITDONE, aim_recv_conn_complete, 0);
	aim_conn_addhandler(session, bos_conn, 0x0009, 0x0003, aim_recv_bos_rights, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_ACK, AIM_CB_ACK_ACK, NULL, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_GEN, AIM_CB_GEN_REDIRECT, aim_recv_redirect, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_LOC, AIM_CB_LOC_RIGHTSINFO, aim_recv_locrights, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_BUD, AIM_CB_BUD_RIGHTSINFO, aim_recv_buddy_rights, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_BUD, AIM_CB_BUD_ONCOMING, aim_recv_oncoming, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_BUD, AIM_CB_BUD_OFFGOING, aim_recv_offgoing, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_MSG, AIM_CB_MSG_INCOMING, aim_recv_msg, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_LOC, AIM_CB_LOC_ERROR, aim_recv_err_loc, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_MSG, AIM_CB_MSG_MISSEDCALL, aim_recv_missed, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_MSG, AIM_CB_MSG_CLIENTAUTORESP, aim_recv_clientauto, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_GEN, AIM_CB_GEN_RATECHANGE, aim_recv_rate_change, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_GEN, AIM_CB_GEN_EVIL, aim_recv_evil, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_LOK, AIM_CB_LOK_ERROR, aim_recv_search_error, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_LOK, 0x0003, aim_recv_search_reply, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_MSG, AIM_CB_MSG_ERROR, aim_recv_err_msg, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_MSG, AIM_CB_MSG_MTN, aim_recv_typing, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_LOC, AIM_CB_LOC_USERINFO, aim_recv_userinfo, 0);
	aim_conn_addhandler(session, bos_conn, 0x0004, 0x0005, aim_recv_icbm_param_info, 0);
	aim_conn_addhandler(session, bos_conn, 0x0001, 0x0001, aim_recv_err_other, 0);
	aim_conn_addhandler(session, bos_conn, 0x0003, 0x0001, aim_recv_err_other, 0);
	aim_conn_addhandler(session, bos_conn, 0x0009, 0x0001, aim_recv_err_other, 0);
	aim_conn_addhandler(session, bos_conn, 0x0001, 0x001f, aim_recv_mem_request, 0);
	aim_conn_addhandler(session, bos_conn, 0x0001, 0x000f, aim_recv_selfinfo, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_POP, 0x0002, aim_recv_popup, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_SSI, AIM_CB_SSI_RIGHTSINFO, aim_ssi_recv_rights, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_SSI, AIM_CB_SSI_LIST, aim_ssi_recv_list, 0);
	aim_conn_addhandler(session, bos_conn, AIM_CB_FAM_SSI, AIM_CB_SSI_NOLIST, aim_ssi_recv_list, 0);

	aim_sendcookie(session, bos_conn, authresp->cookie);

	return (1);
}

/*
** XXX: this is wrong, there should be a separate struct
** that holds information about the server, proxy, etc.
*/

int aim_login(struct aim_priv *priv) {
	aim_conn_t *auth_conn;
	char *server;

	screen_status_msg("Logging in as %s...", priv->username);

	if (priv->proxy != NULL) {
		aim_setupproxy(&priv->aim_session, priv->proxy,
			priv->proxy_user, priv->proxy_passwd);
	}

	server = priv->server;
	if (server == NULL)
		server = FAIM_LOGIN_SERVER;

	auth_conn = aim_newconn(&priv->aim_session, AIM_CONN_TYPE_AUTH, server);
	if (auth_conn == NULL) {
		screen_status_msg("Connection error while logging in");
		return (-1);
	}

	if (auth_conn->fd == -1) {
		if (auth_conn->status & AIM_CONN_STATUS_RESOLVERR)
			screen_status_msg("Unable to resolve authorizer hostname");
		else if (auth_conn->status & AIM_CONN_STATUS_CONNERR)
			screen_status_msg("Unable to connect to authorizer");
		else
			screen_status_msg("Error connecting to authorizer");

		aim_conn_kill(&priv->aim_session, &auth_conn);
		return (-1);
	}

	aim_conn_addhandler(&priv->aim_session, auth_conn, 0x0017, 0x0007,
		aim_parse_login, 0);

	aim_conn_addhandler(&priv->aim_session, auth_conn, 0x0017, 0x0003,
		aim_parse_authresp, 0);

	aim_request_login(&priv->aim_session, auth_conn, priv->username);
	return (1);
}

int aim_setup(struct aim_priv *priv) {
	aim_session_t *session = &priv->aim_session;

	memset(session, 0, sizeof(*session));

	aim_session_init(session, AIM_SESS_FLAGS_NONBLOCKCONNECT, 0);
	aim_setdebuggingcb(session, aim_debug);
	aim_tx_setenqueue(session, AIM_TX_IMMEDIATE, NULL);

	session->aux_data = priv;

	return (0);
}

int aim_send_msg(struct aim_priv *priv, char *dest, char *msg) {
	return (aim_send_im(&priv->aim_session, dest, 0, msg));
}

int aim_add_block(struct aim_priv *priv, char *target) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_addpord(&priv->aim_session, priv->bos_conn,
			(const char **) &target, 1, AIM_SSI_TYPE_DENY);

	return (ret);
}

int aim_remove_block(struct aim_priv *priv, char *target) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_delpord(&priv->aim_session, priv->bos_conn,
			(const char **) &target, 1, AIM_SSI_TYPE_DENY);

	return (ret);
}

int aim_add_permit(struct aim_priv *priv, char *target) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_addpord(&priv->aim_session, priv->bos_conn,
			(const char **) &target, 1, AIM_SSI_TYPE_PERMIT);

	return (ret);
}

int aim_remove_permit(struct aim_priv *priv, char *target) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_delpord(&priv->aim_session, priv->bos_conn,
			(const char **) &target, 1, AIM_SSI_TYPE_PERMIT);

	return (ret);
}

int aim_group_add(struct aim_priv *priv, struct group *group) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_addgroups(&priv->aim_session, priv->bos_conn,
			(const char **) &group->name, 1);

	return (ret);
}

int aim_group_remove(struct aim_priv *priv, struct group *group) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_delgroups(&priv->aim_session, priv->bos_conn,
			&group->name, 1);

	return (ret);
}

int aim_buddy_add(struct aim_priv *priv, struct buddy *buddy) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_addbuddies(&priv->aim_session, priv->bos_conn,
			buddy->group->name, (const char **) &buddy->nname, 1);

	return (ret);
}

int aim_buddy_remove(struct aim_priv *priv, struct buddy *buddy) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_delbuddies(&priv->aim_session, priv->bos_conn,
			buddy->group->name, &buddy->nname, 1);

	return (ret);
}

int aim_buddy_move(struct aim_priv *priv, struct buddy *buddy, char *old) {
	int ret;

	if (priv->aim_session.ssi.received_data == 0)
		return (-1);

	ret = aim_ssi_movebuddy(&priv->aim_session, priv->bos_conn, old,
			buddy->group->name, buddy->nname);

	return (ret);
}

int aim_warn(struct aim_priv *priv, char *target) {
	screen_status_msg("Warning %s", target);
	return (aim_send_warning(&priv->aim_session, priv->bos_conn, target, 0));
}

int aim_warn_anon(struct aim_priv *priv, char *target) {
	int ret;

	screen_status_msg("Warning %s anonymously", target);
	ret = aim_send_warning(&priv->aim_session, priv->bos_conn,
			target, AIM_WARN_ANON);

	return (ret);
}

int aim_set_away(struct aim_priv *priv, char *away_msg) {
	size_t len;
	int ret;

	if (away_msg == NULL) {
		if (priv->away_msg != NULL)
			free(priv->away_msg);
		priv->away_msg = NULL;
		ret = aim_bos_setprofile(&priv->aim_session,
				priv->bos_conn, NULL, "", pork_caps);

		screen_status_msg("You're no longer away");
		return (ret);
	}

	len = strlen(away_msg);
	if (len > priv->rights.max_away_len) {
		screen_status_msg("Your away message is too long.  The maximum length is %u characters",
			priv->rights.max_away_len);
		return (-1);
	}

	if (priv->away_msg != NULL)
		free(priv->away_msg);
	priv->away_msg = xstrdup(away_msg);

	ret = aim_bos_setprofile(&priv->aim_session,
			priv->bos_conn, NULL, away_msg, pork_caps);

	screen_status_msg("You're now marked as away");
	return (ret);
}

int aim_keepalive(struct aim_priv *priv) {
	return (aim_flap_nop(&priv->aim_session, priv->bos_conn));
}

int aim_set_idle(struct aim_priv *priv, int idle_secs) {
	int ret = aim_bos_setidle(&priv->aim_session, priv->bos_conn, idle_secs);

	if (ret >= 0) {
		if (idle_secs > 0)
			priv->marked_idle = 1;
		else
			priv->marked_idle = 0;
	}

	return (ret);
}

int aim_set_profile(struct aim_priv *priv, char *profile) {
	int ret;	
	size_t len;

	if (profile == NULL) {
		ret = aim_bos_setprofile(&priv->aim_session,
				priv->bos_conn, "", NULL, pork_caps);

		if (priv->profile) {
			free(priv->profile);
			priv->profile = NULL;
		}

		return (ret);
	}

	len = strlen(profile);
	if (len > priv->rights.max_profile_len) {
		screen_status_msg("Your profile is too long.  The maximum length is %u characters",
			priv->rights.max_profile_len);
		return (-1);
	}

	if (priv->profile)
		free(priv->profile);

	priv->profile = xstrdup(profile);
	ret = aim_bos_setprofile(&priv->aim_session,
			priv->bos_conn, profile, NULL, pork_caps);

	return (ret);
}

int aim_get_away_msg(struct aim_priv *priv, char *buddy) {
	int ret;

	ret = aim_getinfo(&priv->aim_session,
			priv->bos_conn, buddy, AIM_GETINFO_AWAYMESSAGE);

	priv->whois_hack = 0;
	return (ret);
}

int aim_whois(struct aim_priv *priv, char *buddy) {
	int ret;

	/*
	** There's no facility provided to us that allows us
	** to fetch a user's profile and away message at the same
	** time.  Instead, we have to request both separately.
	** The 'whois_hack' variable is set so that when the
	** routine that normally would print the away message
	** (the same one that prints profile info) is called
	** it knows only to print the away message and nothing else
	** for this particular request.  Ordinarily, information
	** like signon time, idle time, username, etc is printed
	** when we receive a response to a "get away message"
	** request.
	*/

	ret = aim_getinfo(&priv->aim_session,
			priv->bos_conn, buddy, AIM_GETINFO_GENERALINFO);

	aim_getinfo(&priv->aim_session,
		priv->bos_conn, buddy, AIM_GETINFO_AWAYMESSAGE);

	priv->whois_hack = 1;
	return (ret);
}

void aim_disconnected(struct aim_priv *priv) {
	aim_conn_kill(&priv->aim_session, &priv->bos_conn);

	if (priv->connected) {
		screen_status_msg("The screen name %s has been disconnected",
			priv->username);
	}

	priv->connected = 0;
	priv->flags = 0;
	priv->warn_level = 0;
	priv->idle_time = 0;
	priv->bos_conn = NULL;

	if (opt_get_bool(OPT_AUTO_RECONNECT)) {
		screen_status_msg("Automatically reconnecting");
		aim_login(priv);
	}
}

void aim_kill_all_conn(struct aim_priv *priv) {
	aim_tx_flushqueue(&priv->aim_session);
	aim_session_kill(&priv->aim_session);
}

void *aim_select_all(	dlist_t *list,
						struct timeval *tv,
						int *status,
						aim_conn_t **active_conn)
{
	dlist_t *cur;
	fd_set rfds;
	fd_set wfds;
	int max_fd = 0;
	int ret;

	FD_ZERO(&rfds);
	FD_ZERO(&wfds);

	for (cur = list ; cur != NULL ; cur = cur->next) {
		struct aim_acct *acct = cur->data;
		aim_session_t *session = acct->session;
		aim_conn_t *cur_conn;
		int num_connecting = 0;
		int num_outgoing = 0;

		if (session->connlist == NULL)
			continue;

		if (session->queue_outgoing)
			num_outgoing++;

		cur_conn = session->connlist;
		while (cur_conn != NULL) {
			if (cur_conn->fd == -1) {
				*status = -1;
				*active_conn = cur_conn;
				return (session);
			} else if (cur_conn->status & AIM_CONN_STATUS_INPROGRESS) {
				FD_SET(cur_conn->fd, &wfds);
				num_connecting++;
			}

			FD_SET(cur_conn->fd, &rfds);

			if (cur_conn->fd > max_fd)
				max_fd = cur_conn->fd;

			cur_conn = cur_conn->next;
		}

		if (num_connecting == 0 && num_outgoing > 0) {
			*status = 1;
			*active_conn = NULL;
			return (session);
		}
	}

	ret = select(max_fd + 1, &rfds, &wfds, NULL, tv);
	if (ret == -1 && errno == EINTR) {
		*status = 0;
		*active_conn = NULL;
		return (NULL);
	}

	if (ret >= 1) {
		for (cur = list ; cur != NULL ; cur = cur->next) {
			struct aim_acct *acct = cur->data;
			aim_session_t *session = acct->session;
			aim_conn_t *cur_conn = session->connlist;

			while (cur_conn != NULL) {
				if (FD_ISSET(cur_conn->fd, &rfds) ||
					((cur_conn->status & AIM_CONN_STATUS_INPROGRESS) && FD_ISSET(cur_conn->fd, &wfds)))
				{
					*status = 2;
					*active_conn = cur_conn;
					return (session);
				}
				cur_conn = cur_conn->next;
			}
		}

		*status = 0;
	} else
		*status = ret;

	*active_conn = NULL;
	return (NULL);
}

int aim_send_buddy_list(	aim_session_t *session,
							aim_conn_t *conn,
							struct aim_priv *priv)
{
	dlist_t *cur;
	int ret = 1;

	cur = priv->buddy_pref->group_list;
	while (cur != NULL) {
		struct group *group = cur->data;
		const char **buddies = xmalloc(group->num_members * sizeof(char *));
		size_t num_buddies = 0;
		dlist_t *cur_buddy = group->member_list;
		struct aim_ssi_item *item;

		item = aim_ssi_itemlist_finditem(session->ssi.items, NULL,
				 group->name, AIM_SSI_TYPE_GROUP);
		if (item == NULL)
			aim_ssi_addgroups(session, conn, (const char**) &group->name, 1);

		while (cur_buddy != NULL) {
			struct buddy *buddy = cur_buddy->data;

			item = aim_ssi_itemlist_finditem(session->ssi.items, NULL,
					 buddy->name, AIM_SSI_TYPE_BUDDY);
			if (item == NULL)
				buddies[num_buddies++] = buddy->name;
			cur_buddy = cur_buddy->next;
		}

		ret = aim_ssi_addbuddies(session, conn, group->name,
				buddies, num_buddies);
		free(buddies);

		blist_update_label(priv->blist, group->blist_line);
		cur = cur->next;
	}

	return (ret);
}

int aim_send_permit_list(	aim_session_t *session,
							aim_conn_t *conn,
							struct aim_priv *priv)
{
	int ret = 1;

	if (priv->buddy_pref->permit_list != NULL) {
		dlist_t *cur = priv->buddy_pref->permit_list;
		size_t len = dlist_len(cur);
		const char **permit = xmalloc(len * sizeof(char *));
		size_t num_permit = 0;

		while (cur != NULL) {
			struct aim_ssi_item *item;

			item = aim_ssi_itemlist_finditem(session->ssi.items, NULL,
					cur->data, AIM_SSI_TYPE_PERMIT);
			if (item == NULL)
				permit[num_permit++] = cur->data;
			cur = cur->next;
		}

		ret = aim_ssi_addpord(session, conn, permit,
				num_permit, AIM_SSI_TYPE_PERMIT);
		free(permit);
	}

	return (ret);
}

int aim_send_block_list(	aim_session_t *session,
							aim_conn_t *conn,
							struct aim_priv *priv)
{
	int ret = 1;

	if (priv->buddy_pref->block_list != NULL) {
		dlist_t *cur = priv->buddy_pref->block_list;
		size_t len = dlist_len(cur);
		const char **deny = xmalloc(len * sizeof(char *));
		size_t num_block = 0;

		while (cur != NULL) {
			struct aim_ssi_item *item;

			item = aim_ssi_itemlist_finditem(session->ssi.items, NULL,
					cur->data, AIM_SSI_TYPE_DENY);
			if (item == NULL)
				deny[num_block++] = cur->data;
			cur = cur->next;
		}

		ret = aim_ssi_addpord(session, conn, deny,
				num_block, AIM_SSI_TYPE_DENY);
		free(deny);
	}

	return (ret);
}

int aim_permdeny_mode(struct aim_priv *priv, int mode) {
	if (mode >= 0 && mode <= 5) {
	 	priv->buddy_pref->permdeny_mode = mode;
		aim_ssi_setpermdeny(&priv->aim_session, priv->bos_conn,
			priv->buddy_pref->permdeny_mode, 0xffffffff);
	}

	return (priv->buddy_pref->permdeny_mode);
}

int aim_report_idle(struct aim_priv *priv, int mode) {
	u_int32_t report_idle = aim_ssi_getpresence(priv->aim_session.ssi.items);
	int ret;

	if (mode != 0)
		report_idle |= 0x0000400;
	else
		report_idle &= ~0x0000400;

	priv->buddy_pref->report_idle = ((report_idle & 0x0000400) != 0);
	ret = aim_ssi_setpresence(&priv->aim_session, priv->bos_conn, report_idle);
	return (ret);
}
