/*
** pork_conf.c - pork's configuration parser.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <stdio.h>
#include <ncurses.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/param.h>

#include <pork.h>

static int pork_mkdir(const char *path) {
	struct stat st;

	if (stat(path, &st) != 0) {
		if (mkdir(path, 0700) != 0) {
			screen_status_msg("Error: mkdir %s: %s", path, strerror(errno));
			return (-1);
		}
	} else {
		if (!S_ISDIR(st.st_mode)) {
			screen_status_msg("Error: %s is not a directory.  It should be.",
				path);
			return (-1);
		}
	}

	return (0);
}

static int read_conf(const char *path) {
	FILE *fp;
	char buf[4096];
	u_int32_t line = 0;

	fp = fopen(path, "r");
	if (fp == NULL) {
		if (errno != ENOENT)
			return (-1);
		return (0);
	}

	while (fgets(buf, sizeof(buf), fp) != NULL) {
		char *p;

		++line;

		p = strchr(buf, '\n');
		if (p == NULL) {
			/* XXX - warning: line is too long */
			fclose(fp);
			return (-1);
		}
		*p = '\0';
		p = buf;
		if (*p == '#')
			continue;

		if (*p == opt_get_char(OPT_CMDCHARS))
			p++;

		if (!blank_str(p))
			run_command(p, CMDSET_MAIN);
	}

	fclose(fp);
	return (0);
}

static int read_acct_conf(struct aim_priv *priv, const char *filename) {
	FILE *fp;
	char buf[2048];
	int line = 0;

	fp = fopen(filename, "r");
	if (fp == NULL) {
		if (errno != ENOENT)
			return (-1);

		return (0);
	}

	while (fgets(buf, sizeof(buf), fp) != NULL) {
		char *p;

		++line;

		p = strchr(buf, '\n');
		if (p == NULL) {
			/* XXX - warning: line is too long */
			fclose(fp);
			return (-1);
		}
		*p = '\0';

		p = buf;
		if (*p == '#')
			continue;

		p = strchr(buf, ':');
		if (p == NULL)
			continue;
		*p++ = '\0';

		while (*p == ' ' || *p == '\t')
			p++;

		if (!strcasecmp(buf, "username")) {
			free(priv->username);
			priv->username = xstrdup(p);
		} else if (!strcasecmp(buf, "password")) {
			if (priv->passwd != NULL) {
				memset(priv->passwd, 0, strlen(priv->passwd));
				free(priv->passwd);
			}
			priv->passwd = xstrdup(p);
		} else if (!strcasecmp(buf, "profile")) {
			free(priv->profile);
			priv->profile = xstrdup(p);
		} else {
			screen_status_msg("Error: account config line %d: bad setting: %s",
				buf);
		}
	}

	fclose(fp);
	return (0);
}

static int read_buddy_list(struct aim_priv *priv, const char *filename) {
	FILE *fp;
	char buf[1024];
	u_int32_t line = 0;
	struct group *cur_group = NULL;
	struct buddy_pref *pref = priv->buddy_pref;

	fp = fopen(filename, "r");
	if (fp == NULL) {
		if (errno != ENOENT) {
			screen_status_msg("Can't open buddy list: %s: %s", filename, strerror(errno));
			return (-1);
		}

		return (0);
	}

	pref->report_idle = opt_get_bool(OPT_REPORT_IDLE);
	pref->permdeny_mode = 1;

	while (fgets(buf, sizeof(buf), fp) != NULL) {
		char *p;
		char type;

		++line;
		p = strchr(buf, '\n');
		if (p == NULL) {
			screen_status_msg("Invalid buddy list data at line %u", line);
			fclose(fp);
			return (-1);
		}

		*p = '\0';
		p = buf;

		type = *p++;

		while (*p == ' ' || *p == '\t')
			p++;

		if (type == 'g')
			cur_group = group_add(priv, p, 0);
		else if (type == 'b') {
			char *alias;
			struct buddy *buddy;

			if (cur_group == NULL) {
				screen_status_msg("Invalid buddy list data at line %u", line);
				fclose(fp);
			}
			alias = strchr(p, ':');
			if (alias != NULL)
				*alias++ = '\0';

			buddy = buddy_add(priv, p, cur_group, 0);
			if (alias != NULL)
				buddy_alias(priv, buddy, alias);
		} else if (type == 'p') {
			buddy_add_permit(priv, p, 0);
		} else if (type == 'd') {
			buddy_add_block(priv, p, 0);
		} else if (type == 'm') {
			u_int32_t permdeny_mode;
			char *end;

			permdeny_mode = strtoul(p, &end, 10);
			if (*end != '\0' || permdeny_mode > 5) {
				screen_status_msg(
					"%s: Invalid value for permit/deny mode at line %u",
					p, line);
				fclose(fp);
				return (-1);
			}
			pref->permdeny_mode = permdeny_mode;
		} else if (type == 'i') {
			u_int32_t report_idle_val;
			char *end;

			report_idle_val = strtoul(p, &end, 10);
			if (*end != '\0' || report_idle_val > 1) {
				screen_status_msg(
					"Invalid value for report idle mode at line %u", line);
				fclose(fp);
				return (-1);
			}
			pref->report_idle = report_idle_val;
		} else {
			screen_status_msg("Invalid buddy list data at line %u", line);
			fclose(fp);
			return (-1);
		}
	}

	fclose(fp);
	return (0);
}

int read_user_config(struct aim_priv *priv, const char *user) {
	char nname[NUSER_LEN];
	char buf[PATH_MAX];
	char *pork_dir = opt_get_str(OPT_PORK_DIR);

	if (user == NULL || pork_dir == NULL)
		return (-1);
	normalize(user, nname, sizeof(nname));

	snprintf(buf, sizeof(buf), "%s/%s", pork_dir, nname);
	if (pork_mkdir(buf) != 0)
		return (-1);

	snprintf(buf, sizeof(buf), "%s/%s/logs", pork_dir, nname);
	if (pork_mkdir(buf) != 0)
		return (-1);

	snprintf(buf, sizeof(buf), "%s/%s/buddy_list", pork_dir, nname);
	if (read_buddy_list(priv, buf) != 0)
		screen_status_msg("There was an error reading your buddy list.");

	snprintf(buf, sizeof(buf), "%s/%s/porkrc", pork_dir, nname);
	if (read_conf(buf) != 0)
		screen_status_msg("There was an error reading your porkrc file.");

	snprintf(buf, sizeof(buf), "%s/%s/account", pork_dir, nname);
	if (read_acct_conf(priv, buf) != 0)
		screen_status_msg("Error: Can't read account config file, %s.", buf);

	return (0);
}

static int save_buddy_list(struct aim_priv *priv, const char *filename) {
	char *fn;
	size_t len;
	FILE *fp;
	dlist_t *gcur;
	struct buddy_pref *pref = priv->buddy_pref;

	len = strlen(filename) + sizeof("-TEMP");
	fn = xmalloc(len);
	snprintf(fn, len, "%s-TEMP", filename);

	fp = fopen(fn, "w");
	if (fp == NULL) {
		screen_status_msg("Can't open buddy list file for writing: %s",
			strerror(errno));
		free(fn);
		return (-1);
	}

	fprintf(fp, "m %u\n", pref->permdeny_mode);
	fprintf(fp, "i %u\n", pref->report_idle);

	for (gcur = pref->group_list ; gcur != NULL ; gcur = gcur->next) {
		struct group *gr = gcur->data;
		dlist_t *bcur;

		fprintf(fp, "g %s\n", gr->name);

		for (bcur = gr->member_list ; bcur != NULL ; bcur = bcur->next) {
			struct buddy *buddy = bcur->data;

			fprintf(fp, "b %s:%s\n", buddy->nname, buddy->name);
		}
	}

	for (gcur = pref->permit_list ; gcur != NULL ; gcur = gcur->next)
		fprintf(fp, "p %s\n", (char *) gcur->data);

	for (gcur = pref->block_list ; gcur != NULL ; gcur = gcur->next)
		fprintf(fp, "d %s\n", (char *) gcur->data);

	fchmod(fileno(fp), 0600);
	fclose(fp);

	if (rename(fn, filename) != 0) {
		unlink(fn);
		free(fn);
		return (-1);
	}

	free(fn);
	return (0);
}

static int save_acct_conf(struct aim_priv *priv, char *filename) {
	char *fn;
	size_t len;
	FILE *fp;

	len = strlen(filename) + sizeof("-TEMP");
	fn = xmalloc(len);
	snprintf(fn, len, "%s-TEMP", filename);

	fp = fopen(fn, "w");
	if (fp == NULL) {
		free(fn);
		return (-1);
	}

	if (priv->username != NULL)
		fprintf(fp, "username: %s\n", priv->username);
	if (priv->profile != NULL)
		fprintf(fp, "profile: %s\n", priv->profile);
	if (opt_get_bool(OPT_SAVE_PASSWD) && priv->passwd != NULL)
		fprintf(fp, "password: %s\n", priv->passwd);

	fchmod(fileno(fp), 0600);
	fclose(fp);

	if (rename(fn, filename) != 0) {
		unlink(fn);
		free(fn);
		return (-1);
	}

	free(fn);
	return (0);
}

int save_user_config(struct aim_priv *priv, const char *user) {
	char nname[NUSER_LEN];
	char buf[PATH_MAX];
	char *pork_dir = opt_get_str(OPT_PORK_DIR);

	if (user == NULL || pork_dir == NULL)
		return (-1);
	normalize(user, nname, sizeof(nname));

	snprintf(buf, sizeof(buf), "%s/%s/buddy_list", pork_dir, nname);
	if (save_buddy_list(priv, buf) != 0)
		screen_status_msg("There was an error writing your buddy list.");

	snprintf(buf, sizeof(buf), "%s/%s/account", pork_dir, nname);
	if (save_acct_conf(priv, buf) != 0)
		screen_status_msg("Error: Can't write account config file, %s.", buf);

	return (0);
}

int read_global_config(void) {
	struct passwd *pw;
	char *pork_dir;
	char buf[4096];

	pw = getpwuid(getuid());
	if (pw == NULL)
		return (-1);

	pork_dir = opt_get_str(OPT_PORK_DIR);
	if (pork_dir == NULL) {
		snprintf(buf, sizeof(buf), "%s/.pork", pw->pw_dir);
		opt_set(OPT_PORK_DIR, buf);

		if (pork_mkdir(buf) != 0)
			return (-1);

		pork_dir = opt_get_str(OPT_PORK_DIR);
	}

	snprintf(buf, sizeof(buf), "%s/porkrc", pork_dir);
	return (read_conf(buf));
}

static void write_alias_line(void *data, void *filep) {
	struct alias *alias = data;
	FILE *fp = filep;

	fprintf(fp, "alias %s %s\n", alias->alias, alias->str);
}

static void write_bind_line(void *data, void *filep) {
	struct binding *binding = data;
	FILE *fp = filep;
	char key_name[32];

	bind_get_keyname(binding->key, key_name, sizeof(key_name));
	fprintf(fp, "bind %s %s\n", key_name, binding->binding);
}

static void write_bind_blist_line(void *data, void *filep) {
	struct binding *binding = data;
	FILE *fp = filep;
	char key_name[32];

	bind_get_keyname(binding->key, key_name, sizeof(key_name));
	fprintf(fp, "bind -buddy %s %s\n", key_name, binding->binding);
}

int save_global_config(void) {
	char porkrc[PATH_MAX];
	char *fn;
	size_t len;
	FILE *fp;
	char *pork_dir = opt_get_str(OPT_PORK_DIR);

	if (pork_dir == NULL)
		return (-1);

	snprintf(porkrc, sizeof(porkrc), "%s/porkrc", pork_dir);

	len = strlen(porkrc) + sizeof("-TEMP");
	fn = xmalloc(len);
	snprintf(fn, len, "%s-TEMP", porkrc);

	fp = fopen(fn, "w");
	if (fp == NULL)
		return (-1);

	opt_write(fp);
	fprintf(fp, "\n");
	hash_iterate(&screen.alias_hash, write_alias_line, fp);
	fprintf(fp, "\n");
	hash_iterate(&screen.binds.main.hash, write_bind_line, fp);
	fprintf(fp, "\n");
	hash_iterate(&screen.binds.blist.hash, write_bind_blist_line, fp);

	fclose(fp);

	if (rename(fn, porkrc) != 0) {
		unlink(fn);
		free(fn);
		return (-1);
	}

	free(fn);
	return (0);
}
