/*
** pork_imwindow.c - interface for manipulating conversation/chat/info windows.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>

#include <pork.h>

struct imwindow *imwindow_new(	u_int32_t rows,
								u_int32_t cols,
								u_int32_t refnum,
								u_int32_t type,
								struct aim_priv *priv,
								char *target)
{
	struct imwindow *imwindow;
	WINDOW *swin;
	struct input *input;
	char nname[NUSER_LEN];

	/*
	** There must be at least 3 rows.
	** 1 for output, 1 for status, 1 for input.
	*/

	if (rows < STATUS_ROWS + 1)
		return (NULL);

	/*
	** Use all the rows but those reserved for status
	** for the output window.
	*/
	swin = newwin(rows - STATUS_ROWS, cols, 0, 0);
	if (swin == NULL)
		return (NULL);

	normalize(target, nname, sizeof(nname));

	imwindow = xcalloc(1, sizeof(*imwindow));
	imwindow->refnum = refnum;
	imwindow->type = type;
	imwindow->target = xstrdup(nname);
	imwindow->name = xstrdup(target);
	imwindow->active_binds = &screen.binds.main;
	imwindow->priv = priv;
	imwindow->priv->ref_count++;

	wopt_init(imwindow->opts, target);

	swindow_init(&imwindow->swindow, swin, rows - STATUS_ROWS,
		cols, imwindow->opts);

	/*
	** Use a separate input handler.
	*/

	if (wopt_get_bool(imwindow->opts, WOPT_PRIVATE_INPUT)) {
		input = xmalloc(sizeof(*input));
		input_init(input, cols);
	} else
		input = &screen.input;
	imwindow->input = input;

	if (wopt_get_bool(imwindow->opts, WOPT_SHOW_BUDDY_LIST)) {
		if (priv->blist != NULL)
			blist_show(priv->blist, imwindow);
	}

	return (imwindow);
}

inline void imwindow_rename(struct imwindow *imwindow, char *new_name) {
	free(imwindow->name);
	imwindow->name = new_name;
}

inline void imwindow_resize(struct imwindow *imwindow,
							u_int32_t rows,
							u_int32_t cols)
{
	struct aim_priv *priv = imwindow->priv;

	if (imwindow->blist_visible) {
		if (priv->blist->slist.cols >= cols) {
			imwindow->blist_visible = 0;
			imwindow->active_binds = &screen.binds.main;
			swindow_resize(&imwindow->swindow, rows - STATUS_ROWS, cols);
		} else {
			u_int32_t width = cols - priv->blist->slist.cols;

			swindow_resize(&imwindow->swindow, rows - STATUS_ROWS, width);
		}
	} else
		swindow_resize(&imwindow->swindow, rows - STATUS_ROWS, cols);

	input_resize(imwindow->input, cols);
}

inline int imwindow_refresh_blist(struct imwindow *imwindow) {
	struct aim_priv *priv = imwindow->priv;

	if (imwindow->blist_visible) {
		blist_refresh(priv->blist, imwindow);
		return (0);
	}

	return (-1);
}

inline void imwindow_refresh(struct imwindow *imwindow) {
	if (imwindow->swindow.visible) {
		wnoutrefresh(imwindow->swindow.win);
		imwindow_refresh_blist(imwindow);
		screen_doupdate();
	}
}

void imwindow_destroy(struct imwindow *imwindow) {
	swindow_destroy(&imwindow->swindow);

	if (wopt_get_bool(imwindow->opts, WOPT_PRIVATE_INPUT)) {
		input_destroy(imwindow->input);
		free(imwindow->input);
	}

	imwindow->priv->ref_count--;
	if (imwindow->priv->ref_count == 0) {
		/* XXX: might want to delete it here (config option). */
	}

	free(imwindow->target);
	free(imwindow);
}

void imwindow_switch_focus(struct imwindow *imwindow) {
	if (!imwindow->blist_visible)
		return;

	if (imwindow->input_focus == BINDS_MAIN) {
		imwindow->input_focus = BINDS_BUDDY;
		imwindow->active_binds = &screen.binds.blist;
	} else {
		imwindow->input_focus = BINDS_MAIN;
		imwindow->active_binds = &screen.binds.main;
	}
}

struct imwindow *imwindow_find(const char *target) {
	dlist_t *list_start = screen.window_list;
	dlist_t *cur = list_start;
	char nname[NUSER_LEN];

	normalize(target, nname, sizeof(nname));

	do {
		struct imwindow *imwindow = cur->data;

		if (!strcasecmp(imwindow->target, nname))
			return (imwindow);

		cur = cur->next;
	} while (cur != list_start);

	return (NULL);
}

/*
** Print the instant message that we've just sent
** to the window "win"
*/

void imwindow_send_im(struct imwindow *win, char *str) {
	char *im_send_str = opt_get_str(OPT_FORMAT_IM_SEND);
	char *buf;
	chtype *ch;
	size_t len;
	struct aim_priv *priv = win->priv;

	len = strlen(im_send_str) + strlen(str) + strlen(priv->username) + 1;
	buf = xmalloc(len);
	make_im(im_send_str, buf, len, priv->username, str);
	ch = xmalloc(sizeof(chtype) * len);
	plaintext_to_cstr(buf, ch, len);
	free(buf);

	swindow_input(&win->swindow, ch);
	free(ch);

	if (priv->away_msg != NULL) {
		if (opt_get_bool(OPT_SEND_REMOVES_AWAY))
			aim_set_away(priv, NULL);
	}

	imwindow_refresh(win);
}

/*
** Print the instant message "msg" from "source" to
** the window specified by "win"
*/

void imwindow_recv_im(struct imwindow *win, const char *source, const char *msg) {
	char *im_recv_str = opt_get_str(OPT_FORMAT_IM_RECV);
	char *p;
	size_t len;
	char *buf;
	chtype *ch;

	p = strchr(msg, '\n');
	if (p != NULL)
		*p++ = '\0';

	len = strlen(msg) + strlen(im_recv_str) + strlen(source) + 1;
	buf = xmalloc(len);
	make_im(im_recv_str, buf, len, source, msg);
	ch = xmalloc(sizeof(chtype) * len);

	plaintext_to_cstr(buf, ch, len);
	free(buf);
	swindow_add(&win->swindow, ch, 1);
	free(ch);

	if (p != NULL)
		screen_print_multiline(win, p, 0);

	if (wopt_get_bool(win->opts, WOPT_BEEP_ON_OUTPUT))
		beep();

	imwindow_refresh(win);
}

/*
** Bind the screen name "user" to the current window.
*/

int imwindow_bind_acct(struct imwindow *imwindow, char *user) {
	dlist_t *node;
	struct aim_acct *acct;
	char nname[NUSER_LEN];
	struct aim_priv *old_priv = imwindow->priv;

	normalize(user, nname, sizeof(nname));

	node = dlist_find(screen.acct_list, nname, screen_acct_find);
	if (node == NULL)
		return (-1);
	acct = node->data;

	if (old_priv != acct->priv) {
		imwindow->priv->ref_count--;
		imwindow->priv = acct->priv;
		imwindow->priv->ref_count++;

		if (imwindow->blist_visible) {
			blist_draw(acct->priv->blist);
			blist_refresh(acct->priv->blist, imwindow);
		}
	}

	return (0);
}
